package view;

import javafx.animation.TranslateTransition;
import javafx.scene.Group;
import javafx.scene.shape.*;
import model.*;
import javafx.util.Duration;

import java.util.ArrayList;

public class BoardView {

	public static final int BOARD_SIZE = 16, CENTERED_ORIGIN = -BOARD_SIZE*7/2, BOARD_HEIGHT = -5, GRAVEYARD_HEIGHT = -1;

	private final TileView[] tiles; // contains all the squares/tiles of the board (visual)
	private final PieceView[] pieces; // contains all the 3d pieces (visual)
	private int black_death_count, white_death_count; // graveyard counter
	private PieceView[] blackGraveyard, whiteGraveyard;

	private final Box[] border; // visual border for the board



	public BoardView(Custom3dModel model, CustomMaterial material, Board board) {

		// creates the board squares, and position them properly
		tiles = new TileView[64];
		for(int j = 1 ; j <= 8 ; j++) {
			for(int i = 1 ; i <= 8 ; i++){
				createTile(new Position(i, j));
			}
		}

		// creates the 3d pieces (from the board list)
		pieces = new PieceView[64];
		clean();
		for(int n = 0 ; n < 64 ; n++) {
			Piece piece = board.getPiece(n);
			if(piece != null) {
				createPiece(piece.getPosition(), piece.getColor(), getMeshType(piece), material, model);
			}
		}
		// init graveyards as empty
		black_death_count = 0;
		white_death_count = 0;
		blackGraveyard = new PieceView[16]; // 16 is maximum kills
		whiteGraveyard = new PieceView[16];

		resetColor(material); // init/reset the board squares with black and white

		// creates the board visual border
		border = new Box[4];
		int[] all_x = {0, 1, 0, -1};
		int[] all_y = {1, 0, -1, 0};
		for(int n = 0 ; n < 4 ; n++) {
				border[n] = new Box(BOARD_SIZE/4 + Math.abs(BOARD_SIZE*8.25*all_y[n]), BOARD_SIZE/2*1.3, BOARD_SIZE/4 + Math.abs(BOARD_SIZE*8.25*all_x[n]));
				border[n].setMaterial(material.get(CustomMaterial.WHITE));
				border[n].translateXProperty().set((double) all_x[n] * (BOARD_SIZE*4 + BOARD_SIZE/4/2));
				border[n].translateZProperty().set((double) all_y[n] * (BOARD_SIZE*4 + BOARD_SIZE/4/2));
		}
	}

	public TileView getTile(int id) {
		return tiles[id];
	}

	public Box getTileObj(int id) {
		return tiles[id].getObj();
	}

	public void setTile(TileView tile) {
		tiles[tile.getPosition().getID()] = tile;
	}

	public PieceView getPiece(int id) {
		return pieces[id];
	}

	public MeshView getPieceObj(int id) {
		return pieces[id].getObj();
	}

	public PieceView[] getAllPieces() {
		return pieces;
	}

	public PieceView[] getWhiteGraveyard() {
		return whiteGraveyard;
	}

	public PieceView[] getBlackGraveyard() {
		return blackGraveyard;
	}

	public void setPiece(PieceView piece) {
		pieces[piece.getPosition().getID()] = piece;
	}

	public int getMeshType(Piece piece) {
		int type = Custom3dModel.PAWN; // we can't do switch statement on ".getClass()"
		if (piece.getClass() == Rook.class) { type = Custom3dModel.ROOK; }
		if (piece.getClass() == Knight.class) { type = Custom3dModel.KNIGHT; }
		if (piece.getClass() == Bishop.class) { type = Custom3dModel.BISHOP; }
		if (piece.getClass() == Queen.class) { type = Custom3dModel.QUEEN; }
		if (piece.getClass() == King.class) { type = Custom3dModel.KING; }
		if (piece.getClass() == Nwap.class) { type = Custom3dModel.NWAP; }
		if (piece.getClass() == Magician.class) { type = Custom3dModel.MAGICIAN; }
		if (piece.getClass() == Giant.class) { type = Custom3dModel.GIANT; }
		if (piece.getClass() == Kamikaze.class) { type = Custom3dModel.KNIGHT; }
		return type;
	}

	private void createTile(Position position) {
		// creates a tile and with the right size and position
		Box box = new Box(BOARD_SIZE, (double) BOARD_SIZE / 2, BOARD_SIZE);
		box.setTranslateX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		box.setTranslateZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		// sets the tile on the list
		setTile(new TileView(box, position));
	}

	public void createPiece(Position position, int color, int type, CustomMaterial material, Custom3dModel model) {
		// creates a piece, 3d-model, materials, draw-mode, scale, and the right height (Y-axis)
		MeshView obj = new MeshView(model.get(type));
		obj.setDrawMode(DrawMode.FILL);
		if(color == Piece.WHITE)
			obj.setMaterial(material.get(CustomMaterial.WHITE));
		else
			obj.setMaterial(material.get(CustomMaterial.BLACK));
		int scale = 16; // scales the 3d models to a wanted size, to fit a board tile
		obj.setScaleX(scale);
		obj.setScaleY(-scale); // Y-axis is down (not up) in javaFX, so we need to reverse object
		obj.setScaleZ(-scale); // we reverse Z-axis too, "to keep the orientation of the 3-axis vectors", (reversing Y and Z is equivalent of 180° rotation of X-axis)
		obj.setTranslateY(BOARD_HEIGHT); // place at right 3d height (Y-axis) on board
		// if color is black we rotate
		if(color == Piece.BLACK) {
			obj.setScaleX(-obj.getScaleX());
			obj.setScaleZ(-obj.getScaleZ());
		}
		// sets the piece on the list
		setPiece(new PieceView(obj, position, color));
		// updates piece on the right visual position (corresponding to X-axis and Z-axis placements in 3d)
		updatePiece(position);
	}

	public void updatePiece(Position position) { // updates the piece(i, j), visually, to the right position
		int id = position.getID();
		if(getPiece(id) != null) {
			getPieceObj(id).setTranslateX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
			getPieceObj(id).setTranslateZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		}
	}

	private void updatePieceWithAnimation(Position position) { // updates the piece(i, j), visually, to the right position
		int id = position.getID();
		MeshView obj = getPieceObj(id);
		if(getPiece(id) != null) {
			TranslateTransition animation = new TranslateTransition();
			animation.setDuration(Duration.millis(300));
			animation.setNode(obj);
			animation.setFromX(obj.getTranslateX());
			animation.setFromZ(obj.getTranslateZ());
			animation.setToX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
			animation.setToZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
			animation.play();
		}
	}

	private void updateGraveyardPiece(PieceView piece) { // updates position (visual) of a dead piece
		int x = piece.getPosition().getX();
		int y = piece.getPosition().getY();
		if(x == Position.WHITE_GRAVEYARD) {
			piece.getObj().setTranslateX(-2 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(y * BOARD_SIZE*0.65f - BOARD_SIZE*5);
		} else { // x == Position.BLACK_GRAVEYARD
			piece.getObj().setTranslateX(9 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(-y * BOARD_SIZE*0.65f + BOARD_SIZE*5);
		}
		piece.getObj().setTranslateY(GRAVEYARD_HEIGHT); // place at right height (Y axis) on table
	}

	public void resetColor(CustomMaterial material) {
		for(int n = 0, j = 1 ; j <= 8 ; j++) {
			for (int i = 1; i <= 8; i++, n++) {
				// resets board with black and white - bottom left (first case) is black, then alternates
				if((i + j) % 2 == 0) {
					getTileObj(n).setMaterial(material.get(CustomMaterial.BLACK_TILE)); // (visually)
				} else {
					getTileObj(n).setMaterial(material.get(CustomMaterial.WHITE_TILE)); // (visually)
				}
				getTile(n).setState(TileView.NORMAL); // (on 'state' attribute)
				if(getPiece(n) != null) {
					getPiece(n).resetTargeted(material); // reset the 3d pieces color too (visual)
				}
			}
		}
	}

	public void setColor(CustomMaterial material, Position position, int type) {
		int id = position.getID();
		// set the new color-type (visually)
		getTileObj(id).setMaterial(material.get(type));
		// save color-type in the "state" attribute of that tile
		switch(type) {
			case CustomMaterial.EXPLOSION_TILE -> getTile(id).setState(TileView.EXPLOSION);
			case CustomMaterial.SELECTED_TILE -> getTile(id).setState(TileView.SELECTED);
			case CustomMaterial.MOVABLE_TILE -> getTile(id).setState(TileView.MOVABLE);
			case CustomMaterial.ATTACK_TILE -> getTile(id).setState(TileView.ATTACK);
			case CustomMaterial.SPECIAL_TILE -> getTile(id).setState(TileView.SPECIAL);
			default -> getTile(id).setState(TileView.NORMAL);
		}
	}

	public void movePiece(Position origin, Position arrival) {
		PieceView piece = getPiece(origin.getID());
		piece.setPosition(arrival); // changes position to arrival
		setPiece(piece); // copies the piece from origin ID to arrival ID
		pieces[origin.getID()] = null; // removes the origin ID piece
		// updatePiece(arrival); // updates 3d model position (visual)
		updatePieceWithAnimation(arrival); // updates 3d model position (visual) but animated
	}

	// places a piece on graveyard
	public void killPiece(Position p) { // graveyards are specified as negative x
		PieceView piece = getPiece(p.getID());
		if(piece.getColor() == Piece.WHITE){
			whiteGraveyard[white_death_count] = piece;
			white_death_count++;
			piece.setPosition(new Position(1, white_death_count));
			piece.getPosition().setX(Position.WHITE_GRAVEYARD);
		} else {
			blackGraveyard[black_death_count] = piece;
			black_death_count++;
			piece.setPosition(new Position(1, black_death_count));
			piece.getPosition().setX(Position.BLACK_GRAVEYARD);
		}
		updateGraveyardPiece(piece);
		pieces[p.getID()] = null;
	}

	public void mutationPiece(Custom3dModel model, Position selected, int type) {
		// for view side, we can't "overwrite" the piece, we need to "edit" it,
		// otherwise the old 3d model would stay on the scene
		MeshView obj = getPieceObj(selected.getID());
		obj.setMesh(model.get(type));
	}

	public void addAllToScene(Group group) {
		for(int n = 0 ; n < 64 ; n++) {
			if(getTile(n) != null) {
				group.getChildren().add(getTileObj(n));
			}
		}
		for(int n = 0 ; n < 4 ; n++) {
			group.getChildren().add(border[n]);
		}
		addPiecesOnlyToScene(group);
	}

	public void addPiecesOnlyToScene(Group group) {
		for(int n = 0 ; n < 64 ; n++) {
			if (getPiece(n) != null) {
				group.getChildren().add(getPieceObj(n));
			}
		}
	}

	public void refresh(Board b, Custom3dModel model) { // refresh the 3d objects/meshes, from piece type
		for(int j = 1 ; j <= 8 ; j++) {
			for (int i = 1; i <= 8; i++) {
				Position p = new Position(i, j);
				Piece piece = b.getPiece(p);
				if (piece != null) {
					MeshView obj = getPieceObj(p.getID());
					TriangleMesh newMesh = model.get(getMeshType(piece)); // get the mesh to use
					obj.setMesh(newMesh); // replaces old mesh by new mesh
				}
			}
		}
	}

	public void visualClean(Group group) { // cleans board visually, then data
		for(int n = 0 ; n < 64 ; n++) {
			PieceView piece = getPiece(n);
			if (piece != null) {
				group.getChildren().remove(piece.getObj());
			}
		}
		clean();
	}

	private void clean() { // private, don't use clean, use visualClean, otherwise meshes will stay on screen
		for(int n = 0 ; n < 64 ; n++) {
			pieces[n] = null;
		}
	}
}
